/* File: drvMM4000.c 		   	*/
/* Version: 1.00		   	*/
/* Date Last Modified: 10/25/98	        */

#include	<vxWorks.h>
#include	<stdioLib.h>
#include	<sysLib.h>
#include	<string.h>
#include	<taskLib.h>
#include        <rngLib.h>
#include	<alarm.h>
#include	<dbDefs.h>
#include	<dbAccess.h>
#include	<fast_lock.h>
#include	<recSup.h>
#include	<devSup.h>
#include        <errMdef.h>
#include	<logLib.h>

#include	"motor.h"
#include	"drvMM4000.h"
#include        "gpibIO.h"
#include        "serialIO.h"

#define STATIC static

#define GPIB_PORT  0
#define RS232_PORT 1

#define READ_RESOLUTION "TU;\r"
#define READ_STATUS     "MS;\r"
#define READ_POSITION   "TP;\r"
#define STOP_ALL        "ST;\r"
#define MOTOR_ON        "MO;\r"
#define REMOTE_MODE     "MR;\r"
#define GET_IDENT       "VE;\r"

#define INPUT_TERMINATOR  '\r'

/* Status byte bits */
#define M_AXIS_MOVING     0x01
#define M_MOTOR_POWER     0x02
#define M_MOTOR_DIRECTION 0x04
#define M_PLUS_LIMIT      0x08
#define M_MINUS_LIMIT     0x10
#define M_HOME_SIGNAL     0x20

#define GPIB_TIMEOUT 2000 /* Command timeout in msec */
#define SERIAL_TIMEOUT 2000 /* Command timeout in msec */

#define BUFF_SIZE 100       /* Maximum length of string to/from MM4000 */

/*----------------debugging-----------------*/
#ifdef	DEBUG
    #define Debug(l, f, args...) { if(l<=drvMM4000debug) printf(f,## args); }
#else
    #define Debug(l, f, args...)
#endif

int MM4000_num_cards = 0;
volatile int drvMM4000debug = 0;
STATIC volatile int motor_scan_rate = SCAN_RATE;

STATIC struct controller **motor_state;

STATIC int total_cards;
STATIC int any_motor_in_motion;
STATIC struct circ_queue mess_queue;	/* in message queue head */
STATIC FAST_LOCK queue_lock;
STATIC struct circ_queue free_list;
STATIC FAST_LOCK freelist_lock;

STATIC SEM_ID motor_sem;

/* This is a temporary fix to introduce a delayed reading of the motor
 * position after a move completes
 */
volatile double drvMM4000ReadbackDelay = 0.;


/*----------------functions-----------------*/
STATIC int recv_mess(int card, char *com, int flag);
STATIC int send_mess(int card, char const *com, char c);
STATIC void start_status(int card);
STATIC int set_status(int card, int signal);
static long report(int level);
static long init();
STATIC int motor_init();
STATIC void query_done(int, int, struct mess_node *);

/*----------------functions-----------------*/

struct driver_table MM4000_access =
{
    motor_init,
    motor_send,
    motor_free,
    motor_card_info,
    motor_axis_info,
    &mess_queue,
    &queue_lock,
    &free_list,
    &freelist_lock,
    &motor_sem,
    &motor_state,
    &total_cards,
    &any_motor_in_motion,
    send_mess,
    recv_mess,
    set_status,
    query_done,
    start_status
};

struct
{
    long number;
#ifdef __cplusplus
    long (*report) (int);
    long (*init) (void);
#else
    DRVSUPFUN report;
    DRVSUPFUN init;
#endif
} drvMM4000 = {2, report, init};


/*********************************************************
 * Print out driver status report
 *********************************************************/
static long report(int level)
{
    int card;

    if (MM4000_num_cards <=0)
	printf("    NO MM4000 controllers found\n");
    else
    {
	for (card = 0; card < MM4000_num_cards; card++)
	{
	    if (motor_state[card])
	    {
		struct mm4000 *cntrl;

		cntrl = (struct mm4000 *) motor_state[card]->mm4000;
		switch (cntrl->port_type)
		{
		case RS232_PORT: 
		    printf("    MM4000 controller %d port type = RS-232, id: %s \n", 
			   card, 
			   motor_state[card]->ident);
		    break;
		case GPIB_PORT:
		    printf("    MM4000 controller %d port type = GPIB, id: %s \n", 
			   card, 
			   motor_state[card]->ident);
		    break;
		default:
		    printf("    MM4000 controller %d port type = Unknown, id: %s \n", 
			   card, 
			   motor_state[card]->ident);
		    break;
		}
	    }
	}
    }
    return (0);
}


static long init()
{
   /* 
    * We cannot call motor_init() here, because that function can do GPIB I/O,
    * and hence requires that the drvGPIB have already been initialized.
    * That cannot be guaranteed, so we need to call motor_init from device
    * support
    */
    /* Check for setup */
    if (MM4000_num_cards <= 0)
    {
	Debug(1, "init(): MM4000 driver disabled. MM4000Setup() missing from startup script.\n");
    }
    return ((long) 0);
}


STATIC void query_done(int card, int axis, struct mess_node *nodeptr)
{
}


/*********************************************************
 * Read the status and position of all motors on a card
 * start_status(int card)
 *            if card == -1 then start all cards
 *********************************************************/
STATIC void start_status(int card)
{
    struct mm4000 *cntrl;
    int itera;

    if (card >= 0)
    {
	cntrl = (struct mm4000 *) motor_state[card]->mm4000;
	send_mess(card, READ_STATUS, NULL);
	recv_mess(card, cntrl->status_string, 1);
	send_mess(card, READ_POSITION, NULL);
	recv_mess(card, cntrl->position_string, 1);
    }
    else
    {
	/* 
	 * For efficiency we send messages to all cards, then read all
	 * responses.  This minimizes the latency due to processing on each card
	 */
	for (itera = 0; (itera < total_cards) && motor_state[itera]; itera++)
	    send_mess(itera, READ_STATUS, NULL);
	for (itera = 0; (itera < total_cards) && motor_state[itera]; itera++)
	{
	    cntrl = (struct mm4000 *) motor_state[itera]->mm4000;
	    recv_mess(itera, cntrl->status_string, 1);
	}
	for (itera = 0; (itera < total_cards) && motor_state[itera]; itera++)
	    send_mess(itera, READ_POSITION, NULL);
	for (itera = 0; (itera < total_cards) && motor_state[itera]; itera++)
	{
	    cntrl = (struct mm4000 *) motor_state[itera]->mm4000;
	    recv_mess(itera, cntrl->position_string, 1);
	}
    }
}


/**************************************************************
 * Parse status and position strings for a card and signal
 * set_status()
 ************************************************************/

STATIC int set_status(int card, int signal)
{
    struct mm4000 *cntrl;
    struct mess_node *nodeptr;
    register struct mess_info *motor_info;
    /* Message parsing variables */
    char *p, *tok_save;
    char buff[BUFF_SIZE];
    int i, pos, status;
    int rtn_state;
    double motorData;

    cntrl = (struct mm4000 *) motor_state[card]->mm4000;
    motor_info = &(motor_state[card]->motor_info[signal]);
    nodeptr = motor_info->motor_motion;

    /* 
     * Parse the status string
     * Status string format: 1MSx,2MSy,3MSz,... where x, y and z are the status
     * bytes for the motors
     */
    pos = signal*5 + 3;	 /* Offset in status string */
    status = cntrl->status_string[pos];
    Debug(5, "set_status(): status byte = %x\n", status);

    if (status & M_MOTOR_DIRECTION)
	motor_info->status |= RA_DIRECTION;
    else
	motor_info->status &= ~RA_DIRECTION;

    if (status & M_AXIS_MOVING)
	motor_info->status &= ~RA_DONE;
    else
    {
	motor_info->status |= RA_DONE;
/* TEMPORARY FIX, Mark Rivers, 2/1/99. The MM4000 has reported that the
 * motor is done moving, which means that the "jerk time" is done.  However,
 * the axis can still be settling.  For now we put in a delay and poll the
 * motor position again. This is not a good long-term solution.
 */
	if (motor_info->pid_present == YES && drvMM4000ReadbackDelay != 0.)
	{
	    taskDelay((int)(drvMM4000ReadbackDelay * sysClkRateGet()));
	    send_mess(card, READ_POSITION, NULL);
	    recv_mess(card, cntrl->position_string, 1);
	}
    }

    if ((status & M_PLUS_LIMIT) || (status & M_MINUS_LIMIT))
	motor_info->status |= RA_OVERTRAVEL;
    else
	motor_info->status &= ~RA_OVERTRAVEL;

    if (status & M_HOME_SIGNAL)
	motor_info->status |= RA_HOME;
    else
	motor_info->status &= ~RA_HOME;

    if (status & M_MOTOR_POWER)
	motor_info->status &= ~EA_POSITION;
    else
	motor_info->status |= EA_POSITION;

    /* encoder status */
    motor_info->status &= ~EA_SLIP;
    motor_info->status &= ~EA_SLIP_STALL;
    motor_info->status &= ~EA_HOME;

    /* 
     * Parse motor position
     * Position string format: 1TP5.012,2TP1.123,3TP-100.567,...
     * Skip to substring for this motor, convert to double
     */

    strcpy(buff, cntrl->position_string);
    tok_save = NULL;
    p = my_strtok(buff, ",", &tok_save);
    for (i=0; i<signal; i++) p = my_strtok(NULL, ",", &tok_save);
    Debug(6, "set_status(): position substring = %s\n", p);
    motorData = atof(p+3);

    if (motorData == motor_info->position)
	motor_info->no_motion_count++;
    else
    {
	motor_info->position = (int32_t) motorData;
	if (motor_state[card]->motor_info[signal].encoder_present == YES)
	    motor_info->encoder_position = (int32_t) motorData;
	else
	    motor_info->encoder_position = 0;

	motor_info->no_motion_count = 0;
    }

    motor_info->status &= ~RA_PROBLEM;

    /* Parse motor velocity? */
    /* NEEDS WORK */

    motor_info->velocity = 0;

    if (!(motor_info->status & RA_DIRECTION))
	motor_info->velocity *= -1;

    rtn_state = (!motor_info->no_motion_count ||
		 (motor_info->status & (RA_OVERTRAVEL | RA_DONE | RA_PROBLEM))) ? 1 : 0;

    /* Test for post-move string. */
    if ((motor_info->status & RA_DONE || motor_info->status & RA_OVERTRAVEL) &&
	 nodeptr != 0 && nodeptr->postmsgptr != 0)
    {
	strcpy(buff, nodeptr->postmsgptr);
	strcat(buff, "\r");
	send_mess(card, buff, NULL);
	nodeptr->postmsgptr = NULL;
    }

    return (rtn_state);
}


/*****************************************************/
/* send a message to the MM4000 board		     */
/* send_mess()			                     */
/*****************************************************/
STATIC int send_mess(int card, char const *com, char inchar)
{
    struct mm4000 *cntrl;

    if (strlen(com) > MAX_MSG_SIZE)
    {
	logMsg((char *) "drvMM4000.cc:send_mess(); message size violation.\n",
	       0, 0, 0, 0, 0, 0);
	return (-1);
    }
    
    if (!motor_state[card])
    {
	logMsg((char *) "drvMM4000.cc:send_mess() - invalid card #%d\n", card,
	       0, 0, 0, 0, 0);
	return (-1);
    }

    if (inchar != (char) NULL)
    {
	logMsg((char *) "drvMM4000.cc:send_mess() - invalid argument = %c\n", inchar,
	       0, 0, 0, 0, 0);
	return (-1);
    }

    Debug(2, "send_mess(): message = %s\n", com);

    cntrl = (struct mm4000 *) motor_state[card]->mm4000;

    switch (cntrl->port_type)
    {
	case GPIB_PORT:
	    gpibIOSend(cntrl->gpibInfo, com, strlen(com), GPIB_TIMEOUT);
	    break;
	case RS232_PORT:
	    serialIOSend(cntrl->serialInfo, com, strlen(com), SERIAL_TIMEOUT);
	    break;
    }
    return (0);
}


/*****************************************************/
/* receive a message from the MM4000 board           */
/* recv_mess()			                     */
/*****************************************************/
STATIC int recv_mess(int card, char *com, int flag)
{
    struct mm4000 *cntrl;
    int timeout;
    int len=0;

    /* Check that card exists */
    if (!motor_state[card])
	return (-1);

    cntrl = (struct mm4000 *) motor_state[card]->mm4000;

    switch (cntrl->port_type)
    {
	case GPIB_PORT:
	    if (flag == FLUSH)
		timeout = 0;
	    else
		timeout	= GPIB_TIMEOUT;
	    len = gpibIORecv(cntrl->gpibInfo, com, BUFF_SIZE, INPUT_TERMINATOR,
			     timeout);
	    break;
	case RS232_PORT:
	    if (flag == FLUSH)
		timeout = 0;
	    else
		timeout	= SERIAL_TIMEOUT;
	    len = serialIORecv(cntrl->serialInfo, com, BUFF_SIZE,
			       INPUT_TERMINATOR, timeout);
	    break;
    }

    if (len == 0)
	com[0] = '\0';
    else
	com[len-1] = '\0';

    Debug(2, "recv_mess(): message = \"%s\"\n", com);
    return (len);
}


/*****************************************************/
/* Setup system configuration                        */
/* MM4000Setup()                                     */
/*****************************************************/
int MM4000Setup(int num_cards,	/* maximum number of controllers in system.  */
	    int num_channels,	/* NOT Used. */
	    int scan_rate)	/* polling rate - 1/60 sec units.  */
{
    int itera;

    if (num_cards < 1 || num_cards > MM4000_NUM_CARDS)
	MM4000_num_cards = MM4000_NUM_CARDS;
    else
	MM4000_num_cards = num_cards;

    /* Set motor polling task rate */
    if (scan_rate >= 1 && scan_rate <= sysClkRateGet())
	motor_scan_rate = sysClkRateGet() / scan_rate;
    else
	motor_scan_rate = SCAN_RATE;

   /* 
    * Allocate space for motor_state structures.  Note this must be done
    * before MM4000Config is called, so it cannot be done in motor_init()
    * This means that we must allocate space for a card without knowing
    * if it really exists, which is not a serious problem
    */
    motor_state = (struct controller **) malloc(MM4000_num_cards *
						sizeof(struct controller *));

    for (itera = 0; itera < MM4000_num_cards; itera++)
	motor_state[itera] = (struct controller *) NULL;

    return (0);
}


/*****************************************************/
/* Configure a controller                            */
/* MM4000Config()                                    */
/*****************************************************/
int MM4000Config(int card,	/* card being configured */
            int port_type,      /* GPIB_PORT or RS232_PORT */
	    int addr1,          /* = link for GPIB or hideos_card for RS-232 */
            int addr2)          /* GPIB address or hideos_task */
{
    struct mm4000 *cntrl;

    if (card < 0 || card >= MM4000_num_cards)
        return (ERROR);

    motor_state[card] = (struct controller *) malloc(sizeof(struct controller));
    motor_state[card]->mm4000 = malloc(sizeof(struct mm4000));
    cntrl = (struct mm4000 *) motor_state[card]->mm4000;

    switch (port_type)
    {
    case GPIB_PORT:
        cntrl->port_type = port_type;
        cntrl->gpib_link = addr1;
        cntrl->gpib_address = addr2;
        break;
    case RS232_PORT:
        cntrl->port_type = port_type;
        cntrl->serial_card = addr1;
        strcpy(cntrl->serial_task, (char *) addr2);
        break;
    default:
        return (ERROR);
    }
    return (0);
}



/*****************************************************/
/* initialize all software and hardware		     */
/* This is called from the initialization routine in */
/* device support.                                   */
/* motor_init()			                     */
/*****************************************************/
STATIC int motor_init()
{
    struct mm4000 *cntrl;
    int itera, index, arg3, arg4;
    char axis_pos[BUFF_SIZE];
    char buff[BUFF_SIZE];
    char *tok_save;
    char *pos_ptr;
    int total_axis = 0;
    int status;
    BOOLEAN errind;

    /* Check for setup */
    if (MM4000_num_cards <= 0)
	return (ERROR);

    tok_save = NULL;

    for (itera = 0; itera < MM4000_num_cards; itera++)
    {
	if (!motor_state[itera])
	    continue;

	total_cards = itera + 1;
	cntrl = (struct mm4000 *) motor_state[itera]->mm4000;

	/* Initialize communications channel */
	errind = OFF;
	switch (cntrl->port_type)
	{
	    case GPIB_PORT:
		cntrl->gpibInfo = gpibIOInit(cntrl->gpib_link,
					     cntrl->gpib_address);
		if (cntrl->gpibInfo == NULL)
		    errind = ON;
		break;
	    case RS232_PORT:
		cntrl->serialInfo = serialIOInit(cntrl->serial_card,
						 cntrl->serial_task);
		if (cntrl->serialInfo == NULL)
		    errind = ON;
		break;
	}

	if (errind == OFF)
	{
	    /* Send a message to the board, see if it exists */
	    /* flush any junk at input port - should not be any data available */
	    do
		recv_mess(itera, buff, FLUSH);
	    while (strlen(buff) != 0);
    
	    send_mess(itera, READ_POSITION, NULL);
	    status = recv_mess(itera, axis_pos, 1);  
	    /* Return value is length of response string */
	}

	if (errind == OFF && status > 0)
	{
	    motor_state[itera]->localaddr = (char *) NULL;
	    motor_state[itera]->motor_in_motion = 0;
	    send_mess(itera, STOP_ALL, NULL);	/* Stop all motors */
            send_mess(itera, REMOTE_MODE, NULL);/* Remote mode - disables front panel */
	    send_mess(itera, GET_IDENT, NULL);	/* Read controller ID string */
	    recv_mess(itera, buff, 1);
	    strcpy(motor_state[itera]->ident, &buff[2]);  /* Skip "VE" */

	    send_mess(itera, READ_POSITION, NULL);
	    recv_mess(itera, axis_pos, 1);

	    /* The return string will tell us how many axes this controller has */
	    for (total_axis = 0, pos_ptr = my_strtok(axis_pos, ",", &tok_save);
		    pos_ptr; pos_ptr = my_strtok(NULL, ",", &tok_save),
		    total_axis++)
		motor_state[itera]->motor_info[total_axis].motor_motion = NULL;

	    motor_state[itera]->total_axis = total_axis;

	    start_status(itera);
	    for (index = 0; index < total_axis; index++)
	    {
		struct mess_info *motor_info = &motor_state[itera]->motor_info[index];
		int loop_state;

		motor_info->status = 0;
		motor_info->no_motion_count = 0;
		motor_info->encoder_position = 0;
		motor_info->position = 0;

                /* Determine if encoder present based on open/closed loop mode. */
		sprintf(buff, "%dTC\r", index + 1);
	        send_mess(itera, buff, NULL);
	        recv_mess(itera, buff, 1);
		loop_state = atoi(&buff[3]);	/* Skip first 3 characters */
		if (loop_state != 0)
		{
                    motor_info->encoder_present = YES;
		    motor_info->status |= EA_PRESENT;
		    motor_info->pid_present = YES;
		    motor_info->status |= GAIN_SUPPORT;
		}

                /* Determine low limit */
                sprintf(buff, "%dTL;\r", index + 1);
	        send_mess(itera, buff, NULL);
	        recv_mess(itera, buff, 1);
                motor_info->low_limit = atof(&buff[3]);/* Skip first 3 characters */

                /* Determine high limit */
                sprintf(buff, "%dTR;\r", index + 1);
	        send_mess(itera, buff, NULL);
	        recv_mess(itera, buff, 1);
                motor_info->high_limit = atof(&buff[3]);	/* Skip first 3 characters */

		set_status(itera, index);  /* Read status of each motor */
	    }
	}
	else
	    motor_state[itera] = (struct controller *) NULL;
    }

    motor_sem = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
    any_motor_in_motion = 0;

    FASTLOCKINIT(&queue_lock);
    FASTLOCK(&queue_lock);
    mess_queue.head = (struct mess_node *) NULL;
    mess_queue.tail = (struct mess_node *) NULL;
    FASTUNLOCK(&queue_lock);

    FASTLOCKINIT(&freelist_lock);
    FASTLOCK(&freelist_lock);
    free_list.head = (struct mess_node *) NULL;
    free_list.tail = (struct mess_node *) NULL;
    FASTUNLOCK(&freelist_lock);

    if (sizeof(int) >= sizeof(char *))
    {
	arg3 = (int) (&MM4000_access);
	arg4 = 0;
    }
    else
    {
	arg3 = (int) ((long) &MM4000_access >> 16);
	arg4 = (int) ((long) &MM4000_access & 0xFFFF);
    }
    taskSpawn((char *) "tmotor", 64, VX_FP_TASK | VX_STDIO, 5000, motor_task,
	      motor_scan_rate, arg3, arg4, 0, 0, 0, 0, 0, 0, 0);
    return (0);
}

/*---------------------------------------------------------------------*/
