/*
File:		$RCSfile$
Version:	$Revision$
Modified By:	$Author$
Last Modified:	$Date$
*/

/*
FILENAME...	motordrvCom.h

USAGE...	This file contains definitions and structures that
		are common to all motor record driver support modules.
*/

#ifndef	INCmotordrvComh
#define	INCmotordrvComh 1

#include <rngLib.h>

/*
Valid command types for the driver. The order is of importance; 0 is
lowest importance.  Device support will choose the greatest one to
use as the driver transaction type.
*/

#define UNDEFINED (unsigned char)0	/* garbage type */
#define IMMEDIATE (unsigned char)1	/* 'i' an execute immediate, no reply */
#define MOVE_TERM (unsigned char)2	/* 't' terminate a previous active motion */
#define MOTION    (unsigned char)3	/* 'm' will produce motion updates */
#define VELOCITY  (unsigned char)4	/* 'v' make motion updates till MOVE_TERM */
#define INFO      (unsigned char)5	/* 'f' get curr motor/encoder pos and stat */
#define QUERY     (unsigned char)6	/* 'q' specialty type, not needed */

/* Macros used to set/clear bits in any_motor_in_motion variable. */
#define SET_MM_ON(v,a)  v|=(1<<a)
#define SET_MM_OFF(v,a) v&=~(1<<a)

/* Misc. defines. */
#define ALL_CARDS -1

#define	FLUSH -1 /* The 3rd argument of driver_table's getmsg() can indicate
		either FLUSH the buffer or the # of commands to process. */

/* message queue management - device and driver support only */
struct mess_node
{
    CALLBACK callback;
    int signal;
    int card;
    unsigned char type;
    char message[MAX_MSG_SIZE];
    long position;
    long encoder_position;
    long velocity;
    unsigned long status;
    struct dbCommon *mrecord;	/* "Hidden" pointer to motor record. */
    struct mess_node *next;
    char *postmsgptr;
    char const *termstring;	/* Termination string for STOP_AXIS command
				    (see process_messages()). */
};

/* initial position query to driver - device and driver support only */
typedef struct mess_card_query
{
    char *card_name;
    char *axis_names;
    int total_axis;
} MOTOR_CARD_QUERY;

typedef struct mess_axis_query
{
    long position;
    long encoder_position;
    unsigned long status;
} MOTOR_AXIS_QUERY;

struct axis_status
{
    char direction;
    char done;
    char overtravel;
    char home;
};

struct encoder_status
{
    char slip_enable;
    char pos_enable;
    char slip_detect;
    char pos_dead;
    char axis_home;
    char unused;
};

struct circ_queue	/* Circular queue structure. */
{
    struct mess_node *head;
    struct mess_node *tail;
};

/*----------------motor state info-----------------*/

struct irqdatastr	/* Used only for VME44. */
{
    /* Interrupt Handling control elements */
    int irqErrno;	/* Error indicator from isr */
    uint8_t irqEnable;
    RING_ID recv_rng;	/* message receiving control */
    SEM_ID recv_sem;
    RING_ID send_rng;	/* message transmitting control */
    SEM_ID send_sem;
};

struct mess_info
{
    struct mess_node *motor_motion;	/* in motion, NULL/node */
    int encoder_present;		/* one YES/NO for each axis */
    int32_t position;		/* one pos for each axis */
    int32_t encoder_position;	/* one pos for each axis */
    int32_t velocity;		/* Raw velocity readback(not implemented) */
    int no_motion_count;
    ULONG status_delay;		/* Insure 10ms delay between motion/velocity
				 * commands and status query. */
    unsigned long status;	/* one pos for each axis */
    int pid_present;		/* PID control indicator for VME58 (YES/NO). */
    double high_limit;		/* MM4000 only; Controller's high travel limit. */
    double low_limit;		/* MM4000 only; Controller's low travel limit. */
};

struct controller	/* Controller board information. */
{
    int motor_in_motion;/* count of motors in motion */
    char ident[50];	/* identification string for this card */
    int total_axis;	/* total axis on this card */
    char *localaddr;	/* address of this card */
    struct irqdatastr *irqdata;	/* VME44 only; IRQ data. */
    void *mm4000;	/* For MM4000 = (struct mm4000 *); otherwise Null. */
    struct mess_info motor_info[MAX_AXIS];
};

/* The "driver_table" structure allows device level access to driver level
   data and functions; it supports "common code" for all motor record device
   drivers.
*/
struct driver_table
{
    int (*init) (void);
    int (*send) (struct mess_node *, struct driver_table *);
    int (*free) (struct mess_node *, struct driver_table *);
    int (*get_card_info) (int, MOTOR_CARD_QUERY *, struct driver_table *);
    int (*get_axis_info) (int, int, MOTOR_AXIS_QUERY *, struct driver_table *);
    struct circ_queue *queptr;
    FAST_LOCK *quelockptr;
    struct circ_queue *freeptr;
    FAST_LOCK *freelockptr;
    SEM_ID *semptr;
    struct controller ***card_array;
    int *cardcnt_ptr;
    int *any_inmotion_ptr;
    int (*sendmsg) (int, char const *, char);
    int (*getmsg) (int, char *, int);
    int (*setstat) (int, int);
    void (*query_done) (int, int, struct mess_node *);
    void (*strtstat) (int);			/* Optionial; start status function or NULL. */
};

/* Function prototypes. */

extern char *my_strtok(char *, char const *, char **);
extern int motor_send(struct mess_node *, struct driver_table *);
extern int motor_free(struct mess_node *, struct driver_table *);
extern int motor_card_info(int, MOTOR_CARD_QUERY *, struct driver_table *);
extern int motor_axis_info(int, int, MOTOR_AXIS_QUERY *, struct driver_table *);
extern int motor_task(int a1, int, int, int, int, int, int, int, int, int a10);

#endif	/* INCmotordrvComh */
